SystemOrganization addCategory: #'Neo-Unicode-Core'!

!Character methodsFor: '*neo-unicode-core' stamp: 'SvenVanCaekenberghe 12/7/2015 23:30'!
unicodeCharacterData
	"Answer the official record from the Unicode Character Database for me"
	
	^ NeoUnicodeCharacterData forCodePoint: self codePoint ! !

PackageManifest subclass: #ManifestNeoUnicodeCore
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!ManifestNeoUnicodeCore commentStamp: '<historical>' prior: 0!
I store metadata for this package. These meta data are used by other tools such as the SmalllintManifestChecker and the critics Browser!

!ManifestNeoUnicodeCore class methodsFor: 'code-critics' stamp: 'HenrikSperreJohansen 12/15/2015 17:09'!
ruleRBStringConcatenationRuleV1FalsePositive
	^ #(#(#(#RGMethodDefinition #(#NeoUnicodeCharacterData #initializeFromHangulSyllable: #false)) #'2015-12-15T17:09:38.796506+01:00') )! !

Object subclass: #NeoUnicodeCaser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeCaser commentStamp: 'SvenVanCaekenberghe 12/8/2015 20:57' prior: 0!
I am NeoUnicodeCaser, a tool to change the case of a string of Unicode characters to lower, upper or title case.

Currently, this implementation is too simplistic.!

!NeoUnicodeCaser methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 20:15'!
case: case from: inputStream to: outputStream
	self assert: (#(uppercase lowercase titelcase) includes: case).
	inputStream do: [ :each |
		| unicodeCharacterData |
		unicodeCharacterData := NeoUnicodeCharacterData forCodePoint: each.
		(unicodeCharacterData perform: case)
			ifNil: [ outputStream nextPut: each ] 
			ifNotNil: [ :value | outputStream nextPut: value ] ]! !

!NeoUnicodeCaser methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 20:16'!
case: case fromCharacters: inputStream to: outputStream
	self assert: (#(uppercase lowercase titelcase) includes: case).
	inputStream do: [ :each |
		(each unicodeCharacterData perform: case)
			ifNil: [ outputStream nextPut: each ] 
			ifNotNil: [ :value | outputStream nextPut: (Character codePoint: value) ] ]! !

!NeoUnicodeCaser methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/8/2015 20:17'!
case: case string: string
	^ String streamContents: [ :out |
			self case: case fromCharacters: string readStream to: out ]! !

Object subclass: #NeoUnicodeCharacterData
	instanceVariableNames: 'codePoint name generalCategory bidirectionalCategory decompositionMapping numericMapping mirrored oldName caseMapping canonicalCombiningClass'
	classVariableNames: 'CompositionExclusions HangulCache JamoShortNames SpecialRanges UnicodeCharacterDatabase'
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeCharacterData commentStamp: 'SvenVanCaekenberghe 12/8/2015 20:43' prior: 0!
I am NeoUnicodeCharacterData, I represent one record in the Unicode Character Database and describe properties of one character.

I am uniquely identified by my code point (#codePoint).

I have an official name (#name). Sometimes I also know my older name (#oldName).

I have a general and a bidirectional category (#generalCategory  #bidirectionalCategory) and I know my canonicalCombiningClasses (#canonicalCombiningClasses).

If applicable, I know my case mapping, what code point is like me, but upper, lower or title cased (#uppercase  #lowercase #titelcase)

If applicable, I know my numerical mapping, number I represent (#decimalDigitValue #digitValue #numericValue).

If I am a precomposed character, I know my decomposition (#decompositionMapping). This information is needed for normalization.

I also known whether I am mirrored (#mirrored).

My class side holds the full database of all official Unicode characters, indexed by codePoint (#database), which is loaded lazily from the official URL.

Given a Unicode code point, my class method #forCodePoint: will answer an instance of me.

The character that I represent is available with #character.  Given a character instance, #unicodeCharacterData will answer an instance of me.

	$a unicodeCharacterData.
	$é unicodeCharacterData.
	$7 unicodeCharacterData.

More:

For the interpretation of these concepts, please consult the Unicode documentation at http://unicode.org

In particular, http://www.unicode.org/reports/tr44/ and the older ftp://ftp.unicode.org/Public/3.0-Update/UnicodeCharacterDatabase-3.0.0.html
!

!NeoUnicodeCharacterData class methodsFor: 'accessing - datasets' stamp: 'SvenVanCaekenberghe 12/17/2015 00:01'!
compositionExclusions
	"D112"
	
	^ CompositionExclusions ifNil: [ CompositionExclusions := self loadCompositionExclusionsNames ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - composition exclusion' stamp: 'SvenVanCaekenberghe 12/16/2015 23:49'!
compositionExclusionsFileReference 
	^ FileLocator imageDirectory / #UCD / 'CompositionExclusions.txt'! !

!NeoUnicodeCharacterData class methodsFor: 'private - composition exclusion' stamp: 'SvenVanCaekenberghe 12/16/2015 23:49'!
compositionExclusionsUrl
	^ 'http://unicode.org/Public/UNIDATA/CompositionExclusions.txt' asZnUrl ! !

!NeoUnicodeCharacterData class methodsFor: 'accessing - datasets' stamp: 'SvenVanCaekenberghe 12/8/2015 21:00'!
database
	"Return the Unicode Character Database an a dictionary mapping each codepoint to an instance of me. Load the database if needed from the official URL over the internet."
	
	UnicodeCharacterDatabase ifNil: [ self initializeDatabase ].
	^ UnicodeCharacterDatabase! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'HenrikSperreJohansen 12/16/2015 14:03'!
extractSpecialRanges: collectionOfCharacterData
	|ranges|
	ranges := (collectionOfCharacterData 
			select: [ :each | (each name endsWith: 'First>') or: [ each name endsWith: 'Last>' ] ])
				pairsCollect: [ :first :last | Array with: first with: last ].
	"Happily, in Pharo, removeAllSuchThat does *not* return the collection of removed elements..."
	collectionOfCharacterData 
			removeAllSuchThat: [ :each | (each name endsWith: 'First>') or: [ each name endsWith: 'Last>' ] ].
	^ranges! !

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/15/2015 17:16'!
forCodePoint: integer
	"Return an instance of me for codepoint integer. Fail if there is no such codepoint."

	^ self
		forCodePoint: integer
		ifAbsent: [ KeyNotFound signalFor: integer in: self database ]! !

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/15/2015 17:13'!
forCodePoint: integer ifAbsent: block
	"Return an instance of me for codepoint integer. Execute block if there is no such codepoint."
	
	^ self database at: integer ifAbsent: [ self hangulDataForCodePoint: integer ifAbsent: block]! !

!NeoUnicodeCharacterData class methodsFor: 'private - composition exclusion' stamp: 'SvenVanCaekenberghe 12/16/2015 23:45'!
getCompositionExclusionsRaw
	^ self compositionExclusionsFileReference exists
		ifTrue: [ 
			self compositionExclusionsFileReference binaryReadStreamDo: [ :bin |
				(ZnCharacterReadStream on: bin) upToEnd ] ]
		ifFalse: [ 
			self compositionExclusionsUrl retrieveContents ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - jamo short names' stamp: 'HenrikSperreJohansen 12/15/2015 16:28'!
getJamoShortNamesRaw
	^ self jamoShortNamesFileReference exists
		ifTrue: [ 
			self jamoShortNamesFileReference binaryReadStreamDo: [ :bin |
				(ZnCharacterReadStream on: bin) upToEnd ] ]
		ifFalse: [ 
			self jamoShortNamesUrl retrieveContents ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'SvenVanCaekenberghe 12/13/2015 17:52'!
getRaw
	^ self unicodeCharacterDatabaseFileReference exists
		ifTrue: [ 
			self unicodeCharacterDatabaseFileReference binaryReadStreamDo: [ :bin |
				(ZnCharacterReadStream on: bin) upToEnd ] ]
		ifFalse: [ 
			self unicodeCharacterDatabaseUrl retrieveContents ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - hangul' stamp: 'HenrikSperreJohansen 12/15/2015 17:12'!
hangulDataForCodePoint: integer
	^ self hangulDataForCodePoint: integer ifAbsent: [ KeyNotFound signalFor: integer in: self database ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - hangul' stamp: 'HenrikSperreJohansen 12/16/2015 14:12'!
hangulDataForCodePoint: integer ifAbsent: aBlock
	"Could use on: DomainError do: aBlock rather than check range here explicitly, not sure of the perf overhead though"

	(integer notNil and: [(16rAC00 <= integer and: [ integer <= 16rD7A3 ])])
		ifFalse: [ ^ aBlock value ].
	^ self hangulDatabase at: integer ifAbsentPut: [ self new initializeFromHangulSyllable: integer ]! !

!NeoUnicodeCharacterData class methodsFor: 'private - hangul' stamp: 'HenrikSperreJohansen 12/15/2015 17:06'!
hangulDatabase 
	"Thumb-in-air figure; assume 1000 hangul characters will be a good working set."
	^HangulCache ifNil: [ HangulCache := LRUCache new maximumWeight: 1000]! !

!NeoUnicodeCharacterData class methodsFor: 'initialization' stamp: 'HenrikSperreJohansen 12/16/2015 14:05'!
initializeDatabase
	| entries |
	entries := self loadUnicodeCharacterDatabase.
	SpecialRanges := self extractSpecialRanges: entries.

	UnicodeCharacterDatabase := IdentityDictionary new: entries size.
	entries do: [ :each | 
		UnicodeCharacterDatabase at: each codePoint put: each ].
	
! !

!NeoUnicodeCharacterData class methodsFor: 'initialization' stamp: 'HenrikSperreJohansen 12/15/2015 17:10'!
initializeJamoShortNames
	JamoShortNames := self loadJamoShortNames ! !

!NeoUnicodeCharacterData class methodsFor: 'accessing - datasets' stamp: 'HenrikSperreJohansen 12/15/2015 17:10'!
jamoShortNames
	JamoShortNames ifNil: [ self initializeJamoShortNames  ].
	^JamoShortNames! !

!NeoUnicodeCharacterData class methodsFor: 'private - jamo short names' stamp: 'HenrikSperreJohansen 12/15/2015 16:28'!
jamoShortNamesFileReference 

	^FileLocator imageDirectory / #UCD / 'Jamo.txt'! !

!NeoUnicodeCharacterData class methodsFor: 'private - jamo short names' stamp: 'HenrikSperreJohansen 12/15/2015 16:29'!
jamoShortNamesUrl
	^'http://unicode.org/Public/UNIDATA/Jamo.txt' asZnUrl ! !

!NeoUnicodeCharacterData class methodsFor: 'private - composition exclusion' stamp: 'SvenVanCaekenberghe 12/16/2015 23:51'!
loadCompositionExclusionsNames
	| compositionExclusions |
	compositionExclusions := IdentitySet new.
	self getCompositionExclusionsRaw linesDo: [ :each | 
		(each notEmpty and: [ each first ~= $# ])
			ifTrue: [ compositionExclusions add: (Integer readFrom: each base: 16) ] ].
	^ compositionExclusions! !

!NeoUnicodeCharacterData class methodsFor: 'private - jamo short names' stamp: 'HenrikSperreJohansen 12/15/2015 16:27'!
loadJamoShortNames
	| shortNames |
	shortNames := IdentityDictionary new.
	self getJamoShortNamesRaw
		linesDo:
			[ :each | 
			(self parseUnicodeJamoShortNames: each)
				ifNotNil: [ :shortName | shortNames add: shortName ] ].
	^ shortNames! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'HenrikSperreJohansen 12/16/2015 13:13'!
loadUnicodeCharacterDatabase
	^ OrderedCollection streamContents: [ :out |
		self getRaw linesDo: [ :each |
			out nextPut: (self parseUnicodeCharacterData: each) ] ]! !

!NeoUnicodeCharacterData class methodsFor: 'queries' stamp: 'SvenVanCaekenberghe 12/17/2015 00:06'!
nonStarterDecompositions
	"D111"
	
	^ self database values select: #isNonStarterDecomposition! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'SvenVanCaekenberghe 12/7/2015 21:46'!
parseUnicodeCharacterData: line
	| fields |
	fields := $; split: line.
	^ self new
		initializeFromFields: fields;
		yourself! !

!NeoUnicodeCharacterData class methodsFor: 'private - jamo short names' stamp: 'HenrikSperreJohansen 12/15/2015 16:28'!
parseUnicodeJamoShortNames: aLine
	| codePoint start end shortName |
	"Don't parse comments and empty lines"
	(aLine isEmpty or: [aLine first = $#]) ifTrue: [ ^nil ].
	"All lines are in format CODEPOINT; SHORTNAME #LONGNAME "
	codePoint := Number readFrom: aLine base: 16.
	start := (aLine indexOf: $;) + 2.
	end := (aLine indexOf: Character space startingAt: start) -1.
	shortName := aLine copyFrom: start to: end.
	^codePoint -> shortName	 ! !

!NeoUnicodeCharacterData class methodsFor: 'queries' stamp: 'SvenVanCaekenberghe 12/17/2015 00:00'!
singletons
	"D110"
	
	^ self database values select: #isSingleton! !

!NeoUnicodeCharacterData class methodsFor: 'accessing - datasets' stamp: 'SvenVanCaekenberghe 12/8/2015 14:13'!
specialRanges
	SpecialRanges ifNil: [ self initializeDatabase ].
	^ SpecialRanges! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'SvenVanCaekenberghe 12/13/2015 19:51'!
unicodeCharacterDatabaseFileReference
	"Decompress http://www.unicode.org/Public/UNIDATA/UCD.zip next to your image"
	
	^ FileLocator imageDirectory / #UCD / 'UnicodeData.txt'! !

!NeoUnicodeCharacterData class methodsFor: 'private - database' stamp: 'SvenVanCaekenberghe 12/11/2015 10:06'!
unicodeCharacterDatabaseUrl
	^ 'http://www.unicode.org/Public/UNIDATA/UnicodeData.txt' asZnUrl! !

!NeoUnicodeCharacterData methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 12/7/2015 23:01'!
= anObject
	self == anObject ifTrue: [ ^ true ].
	self class = anObject class ifFalse: [ ^ false ].
	^ codePoint = anObject codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:42'!
bidirectionalCategory
	"Return the symbol indicating the Bidirection Category of my codepoint"

	^ bidirectionalCategory! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/11/2015 10:37'!
canonicalCombiningClass
	"Return the integer indicating the Canonical Combining Class of my codepoint"

	^ canonicalCombiningClass! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/15/2015 16:58'!
ccc
	^ canonicalCombiningClass ! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:28'!
character
	"Return a character instance corresponding to my codepoint"

	^ Character codePoint: codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:28'!
codePoint
	"Return my codepoint, an integer"
	
	^ codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:32'!
decimalDigitValue
	"Return the decimal digit value of my codepoint as an integer. Could be nil if not applicable."
	
	^ numericMapping ifNotNil: [ numericMapping first ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:45'!
decompositionMapping
	"Return my decomposition mapping if applicable, an array of code points, with an optional symbol tag in the first position. The code points of my decomposition can be combined into my codepoint."
	
	^ decompositionMapping! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:32'!
digitValue
	"Return the digit value of my codepoint as an integer. Could be nil if not applicable."

	^ numericMapping ifNotNil: [ numericMapping second ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:27'!
generalCategory
	"Return the symbol indicating the General Category of my codepoint"

	^ generalCategory! !

!NeoUnicodeCharacterData methodsFor: 'gt-inspector-extension' stamp: 'SvenVanCaekenberghe 12/15/2015 12:03'!
gtInspectorCharacterIn: composite
	<gtInspectorPresentationOrder: 30>
	^ composite table
		title: 'Character';
		display: [
			{
				{ 'codepoint'. self codePoint }.
				{ 'codepoint hex'. 
					String streamContents: [ :stream | 
						stream << 'U+'.
						self codePoint printOn: stream base: 16 nDigits: 4 ].
					self codePoint }.
				{ 'character'. self character printString. self character }.
				{ 'name'. self name }.
				{ 'old name'. self oldName ifNil: [ '' ]. self oldName }.
				{ 'general category'. self generalCategory }.
				{ 'canonical combining class'. self canonicalCombiningClass }.
				{ 'bidirectional category'. self bidirectionalCategory }.
				{ 'mirrored'. self mirrored }.
				{ 'uppercase'. 
					self uppercase ifNotNil: [ :n | Character codePoint: n ] ifNil: [ '' ]. 
					NeoUnicodeCharacterData forCodePoint: self uppercase ifAbsent: [] }.
				{ 'lowercase'. 
					self lowercase ifNotNil: [ :n | Character codePoint: n ] ifNil: [ '' ]. 
					NeoUnicodeCharacterData forCodePoint: self lowercase ifAbsent: [] }.
				{ 'titlecase'. 
					self titlecase ifNotNil: [ :n | Character codePoint: n ] ifNil: [ '' ]. 
					NeoUnicodeCharacterData forCodePoint: self titlecase ifAbsent: [] }.
				{ 'decimal digit value'. 
					self decimalDigitValue ifNil: [ '' ]. 
					self decimalDigitValue }.
				{ 'digit value'. self digitValue ifNil: [ '' ]. self digitValue }.
				{ 'numeric value'. self numericValue ifNil: [ '' ]. self numericValue }.
				{ 'decomposition'. 
					String streamContents: [ :out |
						self decompositionMapping ifNotNil: [ :mapping | | decomposition |
							mapping first isInteger 
								ifTrue: [ decomposition := mapping ]
								ifFalse: [ 
									out << $< << mapping first << $>. 
									decomposition := mapping allButFirst ].
							decomposition 
								do: [ :each | out nextPut: (Character codePoint: each) ]
								separatedBy: [ out << ' + ' ] ] ].
					self decompositionMapping ifNotNil: [ :decomposition | 
						decomposition collect: [ :each | 
							each isInteger ifTrue: [ NeoUnicodeCharacterData forCodePoint: each ] ifFalse: [ each ] ] ] }
				} ];
		column: 'Key' evaluated: #first;
		column: 'Value' evaluated: [ :entry | entry second asString ];
		send: [ :entry | entry at: 3 ifAbsent: [ entry second ] ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 09:50'!
hasCanonicalDecomposition
	^ decompositionMapping notNil and: [ decompositionMapping first isInteger ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 09:51'!
hasStandardCanonicalDecomposition
	^ self hasCanonicalDecomposition and: [ decompositionMapping size = 2 ]! !

!NeoUnicodeCharacterData methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 12/7/2015 23:01'!
hash
	^ codePoint hash! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:52'!
initializeCaseMappingFromFields: fields
	| uppercase lowercase titlecase |
	uppercase := (fields at: 13) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	lowercase := (fields at: 14) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	titlecase := (fields at: 15) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	uppercase notNil | lowercase notNil | titlecase notNil
		ifTrue: [ caseMapping := Array with: uppercase with: lowercase with: titlecase ]! !

!NeoUnicodeCharacterData methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/11/2015 10:35'!
initializeFromFields: fields
	codePoint := Integer readFrom: fields first base: 16.
	name := fields second.
	generalCategory := fields third asSymbol.
	canonicalCombiningClass := fields fourth asNumber.
	bidirectionalCategory := fields fifth asSymbol.
	decompositionMapping := fields sixth ifEmpty: [ ] ifNotEmpty: [ :mapping | self parseDecompositionMapping: mapping ].
	self initializeNumericMappingFromFields: fields.
	mirrored := (fields at: 10) first = $Y.
	oldName := (fields at: 11) ifEmpty: [ ].
	"(fields at: 12) = comment is always empty"
	self initializeCaseMappingFromFields: fields! !

!NeoUnicodeCharacterData methodsFor: 'initialize' stamp: 'HenrikSperreJohansen 12/16/2015 13:14'!
initializeFromHangulSyllable: anInteger 
	"Derived from hangul syllable code point. 
	Ref. http://www.unicode.org/versions/Unicode8.0.0/ch03.pdf section 3.12"
	|  sIndex vIndex tIndex lIndex lPart vPart |
	(16rAC00 <= anInteger and: [ anInteger <= 16rD7A3 ])
		ifFalse: [ DomainError signal: 'Hangul syllable outside range' from: 16rAC01 to: 16rD7A2 ].
	"sBase := 16rAC00.
	lBase := 16r1100.
	vBase := 16r1161.
	tBase := 15r11A7.
	lCount := 19.
	vCount := 21.
	tCount := 28.
	nCount := 588."
	
	codePoint := anInteger.
	sIndex := codePoint - 16rAC00.
	lIndex := sIndex // 588.
	vIndex := sIndex \\ 588 // 28.
	tIndex := sIndex \\ 28.
	lPart := 16r1100 + lIndex.
	vPart := 16r1161 + vIndex.
	decompositionMapping := tIndex = 0 
		ifTrue: [ {lPart . vPart.} ] 
		ifFalse: [ {lPart. vPart. 16r11A7 + tIndex} ] .
	name := decompositionMapping inject: 'HANGUL SYLLABLE ' into: [:sub :next | sub, (self class jamoShortNames at: next) ].
	generalCategory := #Lu.
	bidirectionalCategory := #L.
	canonicalCombiningClass := 0.
	mirrored := false.! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:52'!
initializeNumericMappingFromFields: fields
	| decimalDigitValue digitValue numericValue |
	decimalDigitValue := (fields at: 7) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ]. 
	digitValue := (fields at: 8) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ]. 
	numericValue := (fields at: 9) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ].
	decimalDigitValue notNil | digitValue notNil | numericValue notNil
		ifTrue: [ numericMapping := Array with: decimalDigitValue with: digitValue with: numericValue ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/17/2015 00:06'!
isCompositionExclusion
	^ self class compositionExclusions includes: codePoint ! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/17/2015 00:08'!
isFullCompositionExclusion
	^ self isSingleton or: [ self isNonStarterDecomposition or: [ self isCompositionExclusion  ] ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:33'!
isLetter
	^ generalCategory first = $L! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:48'!
isLowercase
	^ generalCategory = #Ll! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:34'!
isMark
	^ generalCategory first = $M! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/17/2015 00:05'!
isNonStarterDecomposition
	^ self hasCanonicalDecomposition 
			and: [ self isSingleton not
				and: [ 
					self isStarter not 
						or: [ (self class forCodePoint: decompositionMapping first) isStarter not ] ] ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:34'!
isNumber
	^ generalCategory first = $N! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberDecimal
	^ generalCategory = #Nd! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberLetter
	^ generalCategory = #Nl! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberOther
	^ generalCategory = #No! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:35'!
isOther
	^ generalCategory first = $C! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/17/2015 00:08'!
isPrimaryComposite
	^ self isFullCompositionExclusion not! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:34'!
isPunctuation
	^ generalCategory first = $P! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:35'!
isSeparator
	^ generalCategory first = $Z! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/13/2015 17:27'!
isSingleton
	^ self hasCanonicalDecomposition and: [ decompositionMapping size = 1 ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 10:36'!
isStarter
	^ canonicalCombiningClass = 0! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2015 22:34'!
isSymbol
	^ generalCategory first = $S! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:49'!
isTitlecase
	^ generalCategory = #Lt! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:48'!
isUppercase
	^ generalCategory = #Lu! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/15/2015 16:30'!
jamoShortName
	"As the name only exists for the few hangul base characters, they are kept in a separate table, rather than as general properties"
	^self class jamoShortNames at: self codePoint ifAbsent: ''! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
lowercase
	"Return the codepoint of the lowercase version of my codepoint. Could be nil if not applicable"
	
	^ caseMapping ifNotNil: [ caseMapping second ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:38'!
mirrored
	"Return if my codepoint is mirrored or not"
	
	^ mirrored! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:37'!
name
	"Return my official name"
	
	^ name! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:33'!
numericValue
	"Return the numeric value of my codepoint as an integer. Could be nil if not applicable."

	^ numericMapping ifNotNil: [ numericMapping third ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:38'!
oldName
	"Return my old/legacy name, could be nil"
	
	^ oldName! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:44'!
parseDecompositionMapping: string
	^ ((Character space split: string)
			collect: [ :each |
				each first = $<
					ifTrue: [ (each copyFrom: 2 to: each size - 1) asSymbol ]
					ifFalse: [ Integer readFrom: each base: 16 ] ]) asArray! !

!NeoUnicodeCharacterData methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 12/15/2015 10:52'!
printOn: stream
	stream << 'U+'.
	self codePoint printOn: stream base: 16 nDigits: 4.
	stream space; << name.
	self oldName ifNotNil: [ stream space; nextPut: $(; << oldName; nextPut: $) ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
titlecase
	"Return the codepoint of the titlecase version of my codepoint. Could be nil if not applicable"

	^ caseMapping ifNotNil: [ caseMapping third ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
uppercase
	"Return the codepoint of the uppercase version of my codepoint. Could be nil if not applicable"

	^ caseMapping ifNotNil: [ caseMapping first ]! !

Object subclass: #NeoUnicodeNormalizer
	instanceVariableNames: 'combinations combiningCharacters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeNormalizer commentStamp: 'HenrikSperreJohansen 12/14/2015 16:40' prior: 0!
I am NeoUnicodeNormalizer, I implement the normalize operation which transforms Unicode text into an equivalent composed or decomposed form, allowing for easier sorting and searching of text. 

The normalize method supports the standard normalization forms described in http://www.unicode.org/unicode/reports/tr15/tr15-23.html

I will eventually implement the forms NFC, NFD, NFKC and NFKD.

Currently, this implementation is too simplistic.
A compliant  normalizer will know to: 
- Distinguish compatability decomposition mappings (first array entry is symbol)
- Recursively apply appropriate decompositions.
- Apply canonical reordering !

!NeoUnicodeNormalizer class methodsFor: 'utils' stamp: 'SvenVanCaekenberghe 12/15/2015 17:22'!
bubbleSort: sequence from: begin to: end using: sortBlock
	"Sort the elements in sequence from begin to end inclusive using sortBlock in place."
	"https://en.wikipedia.org/wiki/Bubble_sort#Optimizing_bubble_sort"
	"self bubbleSort: #(a 3 2 1 b) copy from: 2 to: 4 using: [ :x :y | x <= y ]"
	
	| count swapped |
	count := end - begin - 1.
	[
		swapped := false.
		begin to: begin + count do: [ :index |
			(sortBlock value: (sequence at: index) value: (sequence at: index + 1))
				ifFalse: [ 
					sequence swap: index with: index + 1. 
					swapped := true ] ].
		count := count - 1 ] doWhileTrue: [ swapped ].
	^ sequence! !

!NeoUnicodeNormalizer class methodsFor: 'utils' stamp: 'SvenVanCaekenberghe 12/15/2015 17:18'!
sortCCC: codePoints
	| data size begin end |
	size := codePoints size.
	data := codePoints collect: [ :each | NeoUnicodeCharacterData forCodePoint: each ].
	begin := 1.
	[ begin < size ] whileTrue: [ 
		[ begin < size and: [ (data at: begin) ccc isZero ] ]
			whileTrue: [ begin := begin + 1 ].
		end := begin.
		[ end < size and: [ (data at: end) ccc isZero not ] ]
			whileTrue: [ end := end + 1 ].
		end = size ifFalse: [ end := end - 1 ].
		self bubbleSort: data from: begin to: end using: [ :x :y | x ccc <= y ccc ].
		begin := end + 1 ].
	^ data collect: #codePoint! !

!NeoUnicodeNormalizer methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/8/2015 12:56'!
combinationOf: first and: second
	^ (second notNil and: [ self isCombiningCharacter: second ])
			ifTrue: [ 
				self combinations 
					at: first 
					ifPresent: [ :compositions | 
						compositions at: second ifAbsent: [ ] ] 
					ifAbsent: [ ] ]! !

!NeoUnicodeNormalizer methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:18'!
combinations
	combinations ifNil: [ self initializeForComposition ].
	^ combinations! !

!NeoUnicodeNormalizer methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:18'!
combiningCharacters
	combiningCharacters ifNil: [ self initializeForComposition ].
	^ combiningCharacters! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/11/2015 22:01'!
composeCharactersCompatibleFrom: inputStream to: outputStream
	"NFKC"
	
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		next := inputStream peek.
		(self combinationOf: current codePoint and: (next ifNotNil: [ next codePoint ]))
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: (Character codePoint: composed) ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/11/2015 22:01'!
composeCharactersFrom: inputStream to: outputStream
	"NFC"
	
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		next := inputStream peek.
		(self combinationOf: current codePoint and: (next ifNotNil: [ next codePoint ]))
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: (Character codePoint: composed) ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/13/2015 21:41'!
composeCompatibleFrom: inputStream to: outputStream
	"NFKC"
	
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		current := self replaceSingleton: current.
		next := inputStream peek.
		(self combinationOf: current and: next)
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: composed ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/13/2015 20:19'!
composeFrom: inputStream to: outputStream
	"NFC"
	
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		current := self replaceSingleton: current.
		next := inputStream peek.
		(self combinationOf: current and: next)
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: composed ] ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/11/2015 23:51'!
composeString: string
	"Return the NFC of string, the canonical composition normal form"
	
	string isByteString ifTrue: [
		"Latin1 strings are already in NFC" 
		^ string ]. 
	^ String streamContents: [ :out |
			self composeCharactersFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/11/2015 23:52'!
composeStringCompatible: string
	"Return the NFKC of string, the compatible composition normal form"
	
	(string isByteString and: [ string isAsciiString ]) ifTrue: [
		"ACII strings are already in NFKC" 
		^ string ]. 
	^ String streamContents: [ :out |
			self composeCharactersCompatibleFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/11/2015 23:13'!
decomposeCharactersCompatibleFrom: inputStream to: outputStream
	"NFKD"
	
	[ inputStream atEnd ] whileFalse: [ 
		| character data |
		character := inputStream next.
		data := NeoUnicodeCharacterData forCodePoint: character codePoint ifAbsent: [ ].
		(data notNil and: [ data hasStandardCanonicalDecomposition ])
			ifTrue: [ 
				data decompositionMapping do: [ :each | 
					outputStream nextPut: (Character codePoint: each) ] ]
			ifFalse: [ 
				outputStream nextPut: character ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/11/2015 23:13'!
decomposeCharactersFrom: inputStream to: outputStream
	"NFD"
	
	[ inputStream atEnd ] whileFalse: [ 
		| character data |
		character := inputStream next.
		data := NeoUnicodeCharacterData forCodePoint: character codePoint ifAbsent: [ ].
		(data notNil and: [ data hasStandardCanonicalDecomposition ])
			ifTrue: [ 
				data decompositionMapping do: [ :each | 
					outputStream nextPut: (Character codePoint: each) ] ]
			ifFalse: [ 
				outputStream nextPut: character ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/13/2015 21:37'!
decomposeCompatibleFrom: inputStream to: outputStream
	"NFKD"
	
	[ inputStream atEnd ] whileFalse: [ 
		| unicodeCharacterData codePoint |
		codePoint := inputStream next.
		unicodeCharacterData := NeoUnicodeCharacterData forCodePoint: codePoint ifAbsent: [ ].
		unicodeCharacterData
			ifNotNil: [
				unicodeCharacterData isSingleton 
					ifTrue: [ outputStream nextPut: unicodeCharacterData decompositionMapping first ]
					ifFalse: [ 
						unicodeCharacterData hasCanonicalDecomposition
							ifTrue: [ outputStream nextPutAll: unicodeCharacterData decompositionMapping ]
							ifFalse: [ 
								unicodeCharacterData decompositionMapping 
									ifNotNil: [ :decomposition | 
										outputStream next: decomposition size - 1 putAll: decomposition startingAt: 2 ]
									ifNil: [ outputStream nextPut: codePoint ] ] ] ]
			ifNil: [ outputStream nextPut: codePoint ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/13/2015 21:31'!
decomposeFrom: inputStream to: outputStream
	"NFD"
	
	[ inputStream atEnd ] whileFalse: [ 
		| unicodeCharacterData codePoint |
		codePoint := inputStream next.
		unicodeCharacterData := NeoUnicodeCharacterData forCodePoint: codePoint ifAbsent: [ ].
		unicodeCharacterData
			ifNotNil: [
				unicodeCharacterData isSingleton 
					ifTrue: [ outputStream nextPut: unicodeCharacterData decompositionMapping first ]
					ifFalse: [ 
						unicodeCharacterData hasCanonicalDecomposition
							ifTrue: [ outputStream nextPutAll: unicodeCharacterData decompositionMapping ]
							ifFalse: [ outputStream nextPut: codePoint ] ] ]
			ifNil: [ outputStream nextPut: codePoint ] ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/11/2015 23:52'!
decomposeString: string
	"Return the NFD of string, the canonical decomposition normal form"
	
	(string isByteString and: [ string isAsciiString ]) ifTrue: [
		"ACII strings are already in NFD" 
		^ string ]. 
	^ String streamContents: [ :out |
			self decomposeCharactersFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/11/2015 23:52'!
decomposeStringCompatible: string
	"Return the NFKD of string, the compatible decomposition normal form"
	
	(string isByteString and: [ string isAsciiString ]) ifTrue: [
		"ACII strings are already in NFKD" 
		^ string ]. 
	^ String streamContents: [ :out |
			self decomposeCharactersCompatibleFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/13/2015 22:19'!
initializeForComposition
	combinations := IdentityDictionary new.
	combiningCharacters := IdentitySet new.
	NeoUnicodeCharacterData database valuesDo: [ :each |
		each hasStandardCanonicalDecomposition ifTrue: [ | combination |
			combination := each decompositionMapping.
			"combination first + combination second = each codePoint"
			combinations 
				at: combination first 
				ifPresent: [ :value | 
					value at: combination second put: each codePoint ]
				ifAbsent: [ | value |
					value := IdentityDictionary new at: combination second put: each codePoint; yourself.
					combinations at: combination first put: value ].
			combiningCharacters add: combination second ] ]! !

!NeoUnicodeNormalizer methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:19'!
isCombiningCharacter: codePoint
	^ self combiningCharacters includes: codePoint! !

!NeoUnicodeNormalizer methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/13/2015 20:19'!
replaceSingleton: codePoint
	| unicodeCharacterData |
	unicodeCharacterData := NeoUnicodeCharacterData forCodePoint: codePoint ifAbsent: [ ^ codePoint ].
	^ unicodeCharacterData isSingleton 
		ifTrue: [ unicodeCharacterData decompositionMapping first ]
		ifFalse: [ codePoint ]! !
